# Copyright (c) 2020-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import warnings
import numpy as np
import pandapower as pp
from packaging import version

from ._pp_bus_to_ls_bus import pp_bus_to_ls
from ._my_const import _MIN_PP_VERSION_ADV_GRID_MODEL

def _aux_add_trafo(converter, model, pp_net, pp_to_ls):
    """
    Add the transformers of the pp_net into the lightsim2grid "model"

    Parameters
    ----------
    converter
    model
    pp_net

    Returns
    -------

    """
    if "parallel" in pp_net.trafo and np.any(pp_net.trafo["parallel"].values != 1):
        raise RuntimeError("Cannot handle 'parallel' trafo columns. Please duplicate the rows if that is the case. "
                           "Some pp_net.trafo[\"parallel\"] != 1 it is not handled by lightsim yet.")

    # fix the missing values
    tap_neutral = 1.0 * pp_net.trafo["tap_neutral"].values
    if np.any(~np.isfinite(tap_neutral)):
        warnings.warn("There were some Nan in the pp_net.trafo[\"tap_neutral\"], they have been replaced by 0")
    tap_neutral[~np.isfinite(tap_neutral)] = 0.

    if np.any(tap_neutral != 0.):
        raise RuntimeError("lightsim converter supposes that tap_neutral is 0 for the transformers")

    tap_step_pct = 1.0 * pp_net.trafo["tap_step_percent"].values
    if np.any(~np.isfinite(tap_step_pct)):
        warnings.warn("There were some Nan in the pp_net.trafo[\"tap_step_percent\"], they have been replaced by 0")
    tap_step_pct[~np.isfinite(tap_step_pct)] = 0.

    tap_pos = 1.0 * pp_net.trafo["tap_pos"].values
    if np.any(~np.isfinite(tap_pos)):
        warnings.warn("There were some Nan in the pp_net.trafo[\"tap_pos\"], they have been replaced by 0")
    tap_pos[~np.isfinite(tap_pos)] = 0.

    shift_ = 1.0 * pp_net.trafo["shift_degree"].values
    if np.any(~np.isfinite(tap_pos)):
        warnings.warn("There were some Nan in the pp_net.trafo[\"shift_degree\"], they have been replaced by 0")
    shift_[~np.isfinite(shift_)] = 0.

    is_tap_hv_side = pp_net.trafo["tap_side"].values == "hv"
    if np.any(~np.isfinite(is_tap_hv_side)):
        warnings.warn("There were some Nan in the pp_net.trafo[\"tap_side\"], they have been replaced by \"hv\"")
    is_tap_hv_side[~np.isfinite(is_tap_hv_side)] = True

    if "tap_phase_shifter" in pp_net.trafo:
        if np.any(pp_net.trafo["tap_phase_shifter"].values):
            raise RuntimeError("Ideal phase shifters are not modeled. Please remove all trafos with "
                               "pp_net.trafo[\"tap_phase_shifter\"] set to True.")
    elif "tap_changer_type" in pp_net.trafo:
        if np.any(pp_net.trafo["tap_changer_type"].values == "Ideal"):
            raise RuntimeError("Ideal phase shifters are not modeled. Please remove all 2-winding trafos "
                               "with \"tap_changer_type\" set to \"Ideal\".")
    elif "tap_changer_type" in pp_net.trafo3w:
        if np.any(pp_net.trafo3w["tap_changer_type"].values == "Ideal"):
            raise RuntimeError("Ideal phase shifters are not modeled. Please remove all 3-winding trafos "
                               "with \"tap_changer_type\" set to \"Ideal\".")

    tap_angles_ = 1.0 * pp_net.trafo["tap_step_degree"].values
    if np.any(~np.isfinite(tap_angles_)):
        warnings.warn("There were some Nan in the pp_net.trafo[\"tap_step_degree\"], they have been replaced by 0")
    tap_angles_[~np.isfinite(tap_angles_)] = 0.
    tap_angles_ = np.deg2rad(tap_angles_)

    # compute physical parameters
    if version.parse(pp.__version__) >= _MIN_PP_VERSION_ADV_GRID_MODEL:
        # TODO
        trafo_r, trafo_x, trafo_b = \
            converter.get_trafo_param(tap_step_pct,
                                      tap_pos,
                                      tap_angles_,  # in radian !
                                      is_tap_hv_side,
                                      pp_net.bus.loc[pp_net.trafo["hv_bus"]]["vn_kv"],
                                      pp_net.bus.loc[pp_net.trafo["lv_bus"]]["vn_kv"],
                                      pp_net.trafo["vk_percent"].values,
                                      pp_net.trafo["vkr_percent"].values,
                                      pp_net.trafo["sn_mva"].values,
                                      pp_net.trafo["pfe_kw"].values,
                                      pp_net.trafo["i0_percent"].values,
                                      )

        # initialize the grid
        model.init_trafo(trafo_r,
                        trafo_x,
                        trafo_b,
                        tap_step_pct,
                        tap_pos,
                        shift_,
                        is_tap_hv_side,
                        pp_bus_to_ls(pp_net.trafo["hv_bus"].values, pp_to_ls),
                        pp_bus_to_ls(pp_net.trafo["lv_bus"].values, pp_to_ls))
    else:
        trafo_r, trafo_x, trafo_b = \
            converter.get_trafo_param_legacy(tap_step_pct,
                                             tap_pos,
                                             tap_angles_,  # in radian !
                                             is_tap_hv_side,
                                             pp_net.bus.loc[pp_net.trafo["hv_bus"]]["vn_kv"],
                                             pp_net.bus.loc[pp_net.trafo["lv_bus"]]["vn_kv"],
                                             pp_net.trafo["vk_percent"].values,
                                             pp_net.trafo["vkr_percent"].values,
                                             pp_net.trafo["sn_mva"].values,
                                             pp_net.trafo["pfe_kw"].values,
                                             pp_net.trafo["i0_percent"].values,
                                             )

        # initialize the grid
        model.init_trafo(trafo_r,
                        trafo_x,
                        trafo_b,
                        tap_step_pct,
                        tap_pos,
                        shift_,
                        is_tap_hv_side,
                        pp_bus_to_ls(pp_net.trafo["hv_bus"].values, pp_to_ls),
                        pp_bus_to_ls(pp_net.trafo["lv_bus"].values, pp_to_ls))

    for tr_id, is_connected in enumerate(pp_net.trafo["in_service"].values):
        if not is_connected:
            # trafo is deactivated
            model.deactivate_trafo(tr_id)



# Copyright (c) 2020-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

__all__ = ["SolverType",
           "ErrorType",
           "AnySolver",  
           "GaussSeidelSolver",
           "GaussSeidelSynchSolver",
           "SparseLUSolver",
           "SparseLUSolverSingleSlack",
           "DCSolver",
           "FDPF_XB_SparseLUSolver",
           "FDPF_BX_SparseLUSolver"]

from lightsim2grid_cpp import SolverType
from lightsim2grid_cpp import ErrorType
from lightsim2grid_cpp import AnySolver
               
from lightsim2grid_cpp import GaussSeidelSolver  # SolverType.GaussSeidel
from lightsim2grid_cpp import GaussSeidelSynchSolver  # SolverType.GaussSeidelSynch
from lightsim2grid_cpp import SparseLUSolver  # SolverType.SparseLU
from lightsim2grid_cpp import SparseLUSolverSingleSlack  # SolverType.SparseLUSingleSlack
from lightsim2grid_cpp import DCSolver  # SolverType.DC
from lightsim2grid_cpp import FDPF_XB_SparseLUSolver  # SolverType.FDPF_XB_SparseLU
from lightsim2grid_cpp import FDPF_BX_SparseLUSolver  # SolverType.FDPF_BX_SparseLU

try:
    from lightsim2grid_cpp import KLUSolver  # SolverType.KLU
    from lightsim2grid_cpp import KLUSolverSingleSlack  # SolverType.KLUSingleSlack
    from lightsim2grid_cpp import KLUDCSolver  # SolverType.KLUDC
    from lightsim2grid_cpp import FDPF_XB_KLUSolver  # SolverType.FDPF_XB_KLU
    from lightsim2grid_cpp import FDPF_BX_KLUSolver  # SolverType.FDPF_BX_KLU
    __all__.append("KLUSolver")
    __all__.append("KLUSolverSingleSlack")
    __all__.append("KLUDCSolver")
    __all__.append("FDPF_XB_KLUSolver")
    __all__.append("FDPF_BX_KLUSolver")
except Exception as exc_:
    # KLU is not available
    pass

try:
    from lightsim2grid_cpp import NICSLUSolver  # SolverType.NICSLU
    from lightsim2grid_cpp import NICSLUSolverSingleSlack  # SolverType.NICSLUSingleSlack
    from lightsim2grid_cpp import NICSLUDCSolver  # SolverType.NICSLUDC
    from lightsim2grid_cpp import FDPF_XB_NICSLUSolver  # SolverType.FDPF_XB_NICSLU
    from lightsim2grid_cpp import FDPF_BX_NICSLUSolver  # SolverType.FDPF_BX_NICSLU
    __all__.append("NICSLUSolver")
    __all__.append("NICSLUSolverSingleSlack")
    __all__.append("NICSLUDCSolver")
    __all__.append("FDPF_XB_NICSLUSolver")
    __all__.append("FDPF_BX_NICSLUSolver")
except Exception as exc_:
    # NICSLU is not available
    pass

try:
    from lightsim2grid_cpp import CKTSOSolver  # SolverType.CKTSO
    from lightsim2grid_cpp import CKTSOSolverSingleSlack  # SolverType.CKTSOSingleSlack
    from lightsim2grid_cpp import CKTSODCSolver  # SolverType.CKTSODC
    from lightsim2grid_cpp import FDPF_XB_CKTSOSolver  # SolverType.FDPF_XB_CKTSO
    from lightsim2grid_cpp import FDPF_BX_CKTSOSolver  # SolverType.FDPF_BX_CKTSO
    __all__.append("CKTSOSolver")
    __all__.append("CKTSOSolverSingleSlack")
    __all__.append("CKTSODCSolver")
    __all__.append("FDPF_XB_CKTSOSolver")
    __all__.append("FDPF_BX_CKTSOSolver")
except Exception as exc_:
    # NICSLU is not available
    pass



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

__all__ = ["GeneratorContainer",
           "GenInfo",
           "SGenContainer",
           "SGenInfo",
           "LoadContainer",
           "LoadInfo",
           "ShuntContainer",
           "ShuntInfo",
           "TrafoContainer",
           "TrafoInfo",
           "LineContainer",
           "LineInfo",
           "DCLineContainer",
           "DCLineInfo",
           ]

from lightsim2grid_cpp import GeneratorContainer
from lightsim2grid_cpp import GenInfo
from lightsim2grid_cpp import SGenContainer
from lightsim2grid_cpp import SGenInfo
from lightsim2grid_cpp import LoadContainer
from lightsim2grid_cpp import LoadInfo
from lightsim2grid_cpp import ShuntContainer
from lightsim2grid_cpp import ShuntInfo
from lightsim2grid_cpp import TrafoContainer
from lightsim2grid_cpp import TrafoInfo
from lightsim2grid_cpp import LineContainer
from lightsim2grid_cpp import LineInfo
from lightsim2grid_cpp import DCLineContainer
from lightsim2grid_cpp import DCLineInfo



# Copyright (c) 2020-2024, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import time
import numpy as np

import grid2op
from grid2op.Reward import BaseReward
from grid2op.Action._backendAction import _BackendAction

from lightsim2grid import LightSimBackend, ContingencyAnalysis
from lightsim2grid.compilation_options import klu_solver_available
from lightsim2grid.solver import SolverType


class N1ContingencyReward(BaseReward):
    """
    This class implements a reward that leverage the :class:`lightsim2grid.ContingencyAnalysis`
    to compute the number of unsafe contingency at any given time.

    Examples
    --------

    This can be used as:

    .. code-block:: python

        import grid2op
        from lightsim2grid.rewards import N1ContingencyReward
        l_ids = [0, 1, 7]
        env = grid2op.make("l2rpn_case14_sandbox",
                           reward_class=N1ContingencyReward(l_ids=l_ids)
                          )
        obs = env.reset()
        obs, reward, *_ = env.step(env.action_space())
        print(f"reward: {reward:.3f}")
        
    """

    def __init__(self,
                 l_ids=None,
                 threshold_margin=1.,
                 dc=False,
                 normalize=False,
                 logger=None,
                 tol=1e-8,
                 nb_iter=10):
        BaseReward.__init__(self, logger=logger)
        self._backend : LightSimBackend = None
        self._backend_action = None
        self._l_ids = None
        self._dc : bool = dc
        self._normalize : bool = normalize
        if l_ids is not None:
            self._l_ids = [int(el) for el in l_ids]
        self._threshold_margin :float = float(threshold_margin)
        if klu_solver_available:
            if self._dc:
                self._solver_type = SolverType.KLUDC
            else:
                self._solver_type = SolverType.KLU
        else:
            if self._dc:
                self._solver_type = SolverType.DC
            else:
                self._solver_type = SolverType.SparseLU
        self._backend_ls = False
        self._tol = tol
        self._nb_iter = nb_iter
        self._timer_call = 0.
        self._timer_pre_proc = 0.
        self._timer_compute = 0.
        self._timer_post_proc = 0.
            
    def initialize(self, env: "grid2op.Environment.Environment"):
        from grid2op.Environment import BaseEnv
        from grid2op.Backend import PandaPowerBackend  # lazy import because grid2op -> pandapower-> lightsim2grid -> grid2op
        if not isinstance(env, BaseEnv):
            raise RuntimeError("You can only initialize this reward with a "
                               "proper grid2op environment (`BaseEnv`)")
             
        if not isinstance(env.backend, (PandaPowerBackend, LightSimBackend)):
            raise RuntimeError("Impossible to use the `N1ContingencyReward` with "
                               "an environment with a backend that is not "
                               "``PandaPowerBackend` nor `LightSimBackend`."
                               )
        if isinstance(env.backend, LightSimBackend):
            self._backend : LightSimBackend = env.backend.copy()
            self._backend_ls : bool  = True
        elif isinstance(env.backend, PandaPowerBackend):
            self._backend = LightSimBackend.init_grid(type(env.backend))()
            self._backend.init_from_loaded_pandapower(env.backend)
            self._backend.is_loaded = True
        else:
            raise NotImplementedError()
        
        self._backend.set_solver_type(self._solver_type)
        conv, exc_ = self._backend.runpf()
        if not conv:
            raise RuntimeError(f"The reward N1ContingencyReward diverge with error {exc_}")
        bk_act_cls = _BackendAction.init_grid(type(env.backend))
        self._backend_action = bk_act_cls()
        if self._l_ids is None:
            self._l_ids = list(range(type(env).n_line))
        
        if len(self._l_ids) == 0:
            raise RuntimeError("Impossible to use the N1ContingencyReward "
                               "without any contingencies !")
        self.reward_min = 0.
        self.reward_max = len(self._l_ids) if not self._normalize else 1.
        # self._contingecy_analyzer = ContingencyAnalysis(self._backend)
        # self._contingecy_analyzer.add_multiple_contingencies(self._l_ids)

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if is_done:
            return self.reward_min
        
        beg = time.perf_counter()
        # retrieve the state of the grid
        self._backend_action.reset()
        act = env.backend.get_action_to_set()
        th_lim_a = 1. * env.get_thermal_limit()
        th_lim_a[th_lim_a <= 1.] = 1.  # assign 1 for the thermal limit
        
        # apply it to the backend
        self._backend_action += act
        self._backend.apply_action(self._backend_action)
        conv, exc_ = self._backend.runpf()
        if not conv:
            self.logger.warn("Cannot set the backend of the `N1ContingencyReward` => divergence")
            return self.reward_min
        
        # synch the contingency analyzer
        contingecy_analyzer = ContingencyAnalysis(self._backend)
        contingecy_analyzer.computer.change_solver(self._solver_type)
        contingecy_analyzer.add_multiple_contingencies(*self._l_ids)
        now_ = time.perf_counter()
        self._timer_pre_proc += now_ - beg
        tmp = contingecy_analyzer.get_flows()
        self.logger.info(f"{contingecy_analyzer.computer.nb_solved()} converging contingencies")
        now_2 = time.perf_counter()
        self._timer_compute += now_2 - now_
        if self._dc:
            # In DC is study p, but take into account q in the limits
            tmp_res = np.abs(tmp[0])  # this is Por
            # now transform the limits in A in MW
            por, qor, vor, aor = env.backend.lines_or_info()
            p_sq = (1e-3 * th_lim_a)**2 * 3. * vor**2 - qor**2
            p_sq[p_sq <= 0.] = 0.
            limits = np.sqrt(p_sq)
        else:
            tmp_res = 1. * tmp[1]
            limits = th_lim_a
        res = ((tmp_res > self._threshold_margin * limits) | (~np.isfinite(tmp_res))).any(axis=1)  # whether one powerline is above its limit, per cont
        res |=  (np.abs(tmp_res) <= self._tol).all(axis=1)  # other type of divergence: all 0.
        res = res.sum()  # count total of n-1 unsafe 
        res = len(self._l_ids) - res  # reward = things to maximise
        if self._normalize:
            res /= len(self._l_ids)
        now_3 = time.perf_counter()
        self._timer_post_proc += now_3 - now_2
        self._timer_call += time.perf_counter() - beg
        return res

    def reset(self, env: "grid2op.Environment.BaseEnv") -> None:
        self._timer_call = 0.
        self._timer_pre_proc = 0.
        self._timer_compute = 0.
        self._timer_post_proc = 0.
        return super().reset(env)
    
    def close(self):
        if self._backend is not None:
            self._backend.close()
        del self._backend
        self._backend = None



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

__all__ = ["TimeSeriesCPP", 
           # deprecated
           "Computers"]

import numpy as np
import warnings


try:
    from grid2op.Chronics import Multifolder, GridStateFromFile
    from lightsim2grid.lightSimBackend import LightSimBackend
    __all__.append("TimeSerie")
    GRID2OP_INSTALLED = True
except ImportError as exc_:
    # grid2Op is not installed
    GRID2OP_INSTALLED = False

from lightsim2grid.solver import SolverType
from lightsim2grid_cpp import TimeSeriesCPP

# deprecated
Computers = TimeSeriesCPP


class ___TimeSerie:
    """
    This helper class, that only works with grid2op when using a LightSimBackend allows to compute
    the flows (at the origin side of the powerline / transformers). It is roughly equivalent to the
    grid2op code:

    .. code-block:: python

        import grid2op
        import numpy as np
        from grid2op.Parameters import Parameters
        from lightsim2grid.LightSimBackend import LightSimBackend

        env_name = ...
        param = Parameters()
        param.NO_OVERFLOW_DISCONNECTION = True

        env = grid2op.make(env_name, param=param, backend=LightSimBackend())

        done = False
        obs = env.reset()
        nb_step = obs.max_step
        Vs = np.zeros((nb_step, 2 * env.n_sub), dtype=complex)
        As = np.zeros((nb_step, env.n_line), dtype=float)
        while not done:
            obs, reward, done, info = env.step(env.action_space())
            Vs[i, :env.n_sub] = env.backend.V
            As[i] = obs.a_or

    Compare to the previous code, it avoid all grid2op code and can be more than 15 times 
    faster (on the case 118).
    
    It also allows to use python threading module, as the c++ computation can be done in different python threads (the GIL is not locked
    during the c++ computation).

    Examples
    ----------

    It can be used as:

    .. code-block:: python

        from lightsim2grid import TimeSerie
        import grid2op
        from lightsim2grid.LightSimBackend import LightSimBackend

        env_name = ...
        env = grid2op.make(env_name, param=param, backend=LightSimBackend())

        time_series = TimeSerie(env)
        res_p, res_a, res_v = time_series.get_flows(scenario_id=..., seed=...)

    """
    def __init__(self, grid2op_env):
        if not GRID2OP_INSTALLED:
            raise RuntimeError("Impossible to use the python wrapper `TimeSerie` "
                               "when grid2op is not installed. Please fall back to the "
                               "c++ version (available in python) with:\n"
                               "\tfrom lightsim2grid.timeSerie import TimeSerieCPP\n"
                               "and refer to the appropriate documentation.")
            
        from grid2op.Environment import Environment  # otherwise i got issues...
        if not isinstance(grid2op_env.backend, LightSimBackend):
            raise RuntimeError("This class only works with LightSimBackend")
        if not isinstance(grid2op_env, Environment):
            raise RuntimeError("Please an environment of class \"Environment\", "
                               "and not \"MultimixEnv\" or \"BaseMultiProcessEnv\"")
        self.grid2op_env = grid2op_env.copy()
        self.computer = TimeSeriesCPP(self.grid2op_env.backend._grid)
        self.prod_p = None
        self.load_p = None
        self.load_q = None
        self.__computed = False
        
        self.available_solvers = self.computer.available_solvers()
        if SolverType.KLU in self.available_solvers:
            # use the faster KLU if available
            self.computer.change_solver(SolverType.KLU)
    
    def get_injections(self, scenario_id=None, seed=None):
        """
        This function allows to retrieve the injection of the given scenario, for the given seed
        from the grid2op internal environment.
        """
        if scenario_id is not None:
            self.grid2op_env.set_id(scenario_id)
        if seed is not None:
            self.grid2op_env.seed(seed)
        obs = self.grid2op_env.reset()
        self.__computed = False
        return self._extract_inj()

    def compute_V_from_inj(self, prod_p, load_p, load_q, v_init=None, ignore_errors=False):
        """
        This function allows to compute the voltages, at each bus given a list of
        productions and loads.

        We do not recommend to use it directly, as the order of the load or generators might vary !
        """

        if len(prod_p.shape) != 2:
            raise RuntimeError("prod_p should be a matrix with rows representing time steps "
                               "and columns representing individual production.")
        if len(load_p.shape) != 2:
            raise RuntimeError("load_p should be a matrix with rows representing time steps "
                               "and columns representing individual loads.")
        if len(load_q.shape) != 2:
            raise RuntimeError("load_q should be a matrix with rows representing time steps "
                               "and columns representing individual loads.")
        if prod_p.shape[0] != load_p.shape[0] or prod_p.shape[0] != load_q.shape[0]:
            raise RuntimeError(f"prod_p, load_p and load_q should have the same number of "
                               f"rows. We found: prod_p.shape[0] = {prod_p.shape[0]}, load_p.shape[0] = {load_p.shape[0]} "
                               f"and load_q.shape[0] = {load_q.shape[0]}")
        if prod_p.shape[1] != self.grid2op_env.n_gen:
            raise RuntimeError(f"The number of generators on the grid {self.grid2op_env.n_gen} "
                               f"is different that the number of columns of the provided prod_p data: "
                               f"prod_p.shape[1] = {prod_p.shape[1]}")
        if load_p.shape[1] != self.grid2op_env.n_load:
            raise RuntimeError(f"The number of loads on the grid {self.grid2op_env.n_load} "
                               f"is different that the number of columns of the provided load_p data: "
                               f"load_p.shape[1] = {load_p.shape[1]}")
        if load_q.shape[1] != self.grid2op_env.n_load:
            raise RuntimeError(f"The number of loads on the grid {self.grid2op_env.n_load} "
                               f"is different that the number of columns of the provided load_q data: "
                               f"load_q.shape[1] = {load_q.shape[1]}")
        if v_init is None:
            v_init_comp = self.grid2op_env.backend.V
        else:
            v_init_comp = 1.0 * v_init  # make a copy !
        status = self.computer.compute_Vs(prod_p,
                                          np.zeros((prod_p.shape[0], 0)),  # no static generators for now !
                                          load_p,
                                          load_q,
                                          v_init_comp,
                                          self.grid2op_env.backend.max_it,
                                          self.grid2op_env.backend.tol)
        if status != 1 and not ignore_errors:
            # raise an error if the powerflow diverged
            raise RuntimeError(f"Some error occurred, the powerflow has diverged after {self.computer.nb_solved()} step(s)")
        elif status != 1:
            # only raise a warning in this case
            warnings.warn(f"Some error occurred, the powerflow has diverged after {self.computer.nb_solved()} step(s)")
        Vs = 1.0 * self.computer.get_voltages()  # If I don't copy, lazy eval may break stuff... 
        # eg test_time_series_dc.py does behave stochastically
        self.__computed = True
        return Vs
        
    def compute_V(self, scenario_id=None, seed=None, v_init=None, ignore_errors=False):
        """
        This function allows to retrieve the complex voltage at each bus of the grid for each step.

        .. warning:: Topology fixed = no maintenance, no attacks, etc.

            As the topology is fixed, this class does not allow to simulate the effect of maintenance or attacks !
        """
        prod_p, load_p, load_q = self.get_injections(scenario_id=scenario_id, seed=seed)
        Vs = self.compute_V_from_inj(prod_p, load_p, load_q, v_init, ignore_errors)
        return Vs

    def compute_A(self):
        """
        This function returns the current flows (in Amps, A) at the origin (for powerline) / high voltage (for transformer) 
        side
        
        It does not recompute the voltages at each buses, it uses the information get from `compute_V` and
        This is why you must call `compute_V(...)` first !
        """
        if not self.__computed:
            raise RuntimeError("This function can only be used if compute_V has been sucessfully called")
        ampss = self.computer.compute_flows()
        return 1000. * ampss

    def compute_P(self):
        """
        This function returns the active power flows (in MW) at the origin (for powerline) / high voltage (for transformer) 
        side
        
        It does not recompute the voltages at each buses, it uses the information get from `compute_V` and
        This is why you must call `compute_V(...)` first !
        """
        if not self.__computed:
            raise RuntimeError("This function can only be used if compute_V has been sucessfully called")
        mws = 1. * self.computer.compute_power_flows() # If I don't copy, lazy eval may break stuff... 
        # eg test_time_series_dc.py does behave stochastically
        return mws

    def get_flows(self, scenario_id=None, seed=None, v_init=None, ignore_errors=False):
        """
        Retrieve the flows for each step simulated.

        Each row of the resulting flow matrix will correspond to a step.

        Examples
        --------

        .. code-block:: python

            import grid2op
            from lightsim2grid import TimeSerie
            from lightsim2grid import LightSimBackend
            env_name = ...
            env = grid2op.make(env_name, backend=LightSimBackend())

            timeserie = TimeSerie(env)
            res_p, res_a, res_v = timeserie.get_flows(scenario_id, seed, v_init, ignore_errors)

            # in this results, then
            # res_a[row_id] will be the flows, on all powerline corresponding to the `row_id` contingency.
            # you can retrieve it with `security_analysis.contingency_order[row_id]`
        """
        
        Vs = self.compute_V(scenario_id, seed, v_init, ignore_errors)
        amps = self.compute_A()
        Ps = self.compute_P()
        return Ps, amps, Vs
    
    def clear(self):
        """
        Clear everything, as if nothing has been computed
        """
        self.computer.clear()
        self.__computed = False
        
        self.prod_p = None
        self.load_p = None
        self.load_q = None
    
    def close(self):
        """permanently close the object"""
        self.grid2op_env.close()
        self.clear()
        self.computer.close()
         
    def _extract_inj(self):
        data_loader = None
        if isinstance(self.grid2op_env.chronics_handler.real_data, Multifolder):
            data_loader = self.grid2op_env.chronics_handler.real_data.data
        else:
            data_loader = self.grid2op_env.chronics_handler.data

        if not isinstance(data_loader, GridStateFromFile):
            raise RuntimeError("This function only work with chronics coming from files at the moment")
        
        self.prod_p = 1.0 * data_loader.prod_p
        self.load_p = 1.0 * data_loader.load_p
        self.load_q = 1.0 * data_loader.load_q
        return self.prod_p, self.load_p, self.load_q


if GRID2OP_INSTALLED:
    TimeSerie = ___TimeSerie
    


# Copyright (c) 2024, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

from packaging import version
# pandapower version with more advanced grid modelling
_MIN_PP_VERSION_ADV_GRID_MODEL = version.parse("2.15")  



# Copyright (c) 2020-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import warnings
import numpy as np


def _aux_check_legit(pp_net):
    """
    Check that all the dataframes of the pandapower network can be handled in lightsim2grid.

    Parameters
    ----------
    pp_net

    Returns
    -------

    """

    if "trafo3w" in pp_net and pp_net.trafo3w.shape[0]:
        raise RuntimeError("Unsupported element found (Three Winding Transformer - \"pp_net.trafo3w\") "
                           "in pandapower network")
    if "switch" in pp_net and pp_net.switch.shape[0]:
        warnings.warn("There are switches on the pandapower network, they will not be used in lightsim2grid.")
    if "motor" in pp_net and pp_net.motor.shape[0]:
        raise RuntimeError("Unsupported element found (motor - \"pp_net.motor\") in pandapower network")
    if "asymmetric_load" in pp_net and pp_net.asymmetric_load.shape[0]:
        raise RuntimeError("Unsupported element found (Asymmetric Load - \"pp_net.asymmetric_load\") "
                           "in pandapower network")
    if "impedance" in pp_net and pp_net.impedance.shape[0]:
        raise RuntimeError("Unsupported element found (Impedance - \"pp_net.impedance\") "
                           "in pandapower network")
    if "ward" in pp_net and pp_net.ward.shape[0]:
        raise RuntimeError("Unsupported element found (Ward - \"pp_net.ward\") "
                           "in pandapower network")
    if "xward" in pp_net and pp_net.xward.shape[0]:
        raise RuntimeError("Unsupported element found (Extended Ward - \"pp_net.xward\") "
                           "in pandapower network")
    # supported now
    # if "dcline" in pp_net and pp_net.dcline.shape[0]:
    #     raise RuntimeError("Unsupported element found (DC Line - \"pp_net.dcline\") "
    #                        "in pandapower network")

    # bus indexes should start at 0 and be contiguous
    # if np.any(np.sort(pp_net.bus.index) != np.arange(pp_net.bus.shape[0])):
    #     raise RuntimeError("In order to work, pandapower bus indexes should start at 0 and be contiguous. "
    #                        "Make sure that `pp_net.bus.index` have this property. You can write a github "
    #                        "issue if you want improvment on this regard.")
    
    if "_options" in pp_net and \
       "trafo_model" in pp_net["_options"] and \
       pp_net["_options"]["trafo_model"] != "t":
        raise RuntimeError("Cannot handle pp_net with trafo not modeled in \"t\". Please change "
                           "your network and load it again.")



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import warnings

import numpy as np
from ._aux_add_sgen import SOME_KIND_OF_INF_FOR_PMIN_PMAX
from ._pp_bus_to_ls_bus import pp_bus_to_ls


def _aux_add_dc_line(model, pp_net, pp_to_ls):
    """
    Add the transformers of the pp_net into the lightsim2grid "model"

    Parameters
    ----------
    converter
    model
    pp_net

    Returns
    -------

    """
    if "parallel" in pp_net.dcline and np.any(pp_net.dcline["parallel"].values != 1):
        raise RuntimeError("Cannot handle 'parallel' dcline columns. Please duplicate the rows if that is the case. "
                           "Some pp_net.dcline[\"parallel\"] != 1 it is not handled by lightsim yet.")

    if pp_net.dcline.shape[0] == 0:
        # nothing to do if no dc line
        return
    
    branch_from_id = pp_bus_to_ls(pp_net.dcline["from_bus"].values, pp_to_ls)
    branch_to_id = pp_bus_to_ls(pp_net.dcline["to_bus"].values, pp_to_ls)
    p_mw = -pp_net.dcline["p_mw"].values
    if np.any(~np.isfinite(p_mw)):
        warnings.warn("Some non finite values are found for p_mw, they have been replaced by 0.")
        p_mw[~np.isfinite(p_mw)] = 0.
        
    loss_percent = pp_net.dcline["loss_percent"].values
    loss_mw = pp_net.dcline["loss_mw"].values
    vm_or_pu = pp_net.dcline["vm_from_pu"].values
    vm_ex_pu = pp_net.dcline["vm_to_pu"].values
    
    min_q_or = pp_net.dcline["min_q_from_mvar"].values
    if np.any(~np.isfinite(min_q_or)):
        min_q_or[~np.isfinite(min_q_or)] = -SOME_KIND_OF_INF_FOR_PMIN_PMAX
        
    max_q_or = pp_net.dcline["max_q_from_mvar"].values
    if np.any(~np.isfinite(max_q_or)):
        max_q_or[~np.isfinite(max_q_or)] = SOME_KIND_OF_INF_FOR_PMIN_PMAX
        
    min_q_ex = pp_net.dcline["min_q_to_mvar"].values
    if np.any(~np.isfinite(min_q_ex)):
        min_q_ex[~np.isfinite(min_q_ex)] = -SOME_KIND_OF_INF_FOR_PMIN_PMAX
        
    max_q_ex = pp_net.dcline["max_q_to_mvar"].values
    if np.any(~np.isfinite(max_q_ex)):
        max_q_ex[~np.isfinite(max_q_ex)] = +SOME_KIND_OF_INF_FOR_PMIN_PMAX
                          
    model.init_dclines(branch_from_id.astype(np.int32),
                       branch_to_id.astype(np.int32),
                       p_mw,
                       loss_percent,
                       loss_mw,
                       vm_or_pu,
                       vm_ex_pu,
                       min_q_or,
                       max_q_or,
                       min_q_ex,
                       max_q_ex
                       )
    
    for dcl_id, is_connected in enumerate(pp_net.dcline["in_service"].values):
        if not is_connected:
            # trafo is deactivated
            model.deactivate_dcline(dcl_id)



# Copyright (c) 2023-2024, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import warnings
import copy
import numpy as np
import pandas as pd
import pypowsybl as pypo
from typing import Optional, Union
from packaging import version

PP_BUG_RATIO_TAP_CHANGER = version.parse("1.9")
PYPOWSYBL_VER = version.parse(pypo.__version__)
from lightsim2grid_cpp import GridModel


def _aux_get_bus(bus_df, df, conn_key="connected", bus_key="bus_id"):
    if df.shape[0] == 0:
        # no element of this type so no problem
        return np.zeros(0, dtype=int), np.ones(0, dtype=bool)
    # retrieve which elements are disconnected 
    mask_disco = ~df[conn_key]
    if mask_disco.all():
        raise RuntimeError("All element of the same type are disconnected, the init will not work.")
    first_el_co = np.where(~mask_disco.values)[0][0]
    # retrieve the bus where the element are
    tmp_bus_id = df[bus_key].copy()
    tmp_bus_id[mask_disco] = df.iloc[first_el_co][bus_key]  # assign a "random" bus to disco element
    bus_id = bus_df.loc[tmp_bus_id.values]["bus_id"].values
    # deactivate the element not on the main component
    # wrong_component = bus_df.loc[tmp_bus_id.values]["connected_component"].values != 0
    # mask_disco[wrong_component] = True
    # assign bus -1 to disconnected elements
    bus_id[mask_disco] = -1
    return bus_id, mask_disco.values


def init(net : pypo.network.Network,
         gen_slack_id: Union[int, str] = None,
         slack_bus_id: int = None,
         sn_mva = 100.,
         sort_index=True,
         f_hz = 50.,  # unused
         net_pu : Optional[pypo.network.Network] = None,
         only_main_component=True,
         return_sub_id=False,
         n_busbar_per_sub=None,  # new in 0.9.1
         buses_for_sub=None,  # new in 0.9.1
         ):
    model = GridModel()
    # model.set_f_hz(f_hz)
    
    if gen_slack_id is not None and slack_bus_id is not None:
        raise RuntimeError("Impossible to intialize a grid with both gen_slack_id and slack_bus_id")
    
    # assign unique id to the buses
    bus_df_orig = net.get_buses()
    if sort_index:
        bus_df = bus_df_orig.sort_index()
    else:
        bus_df = bus_df_orig
    
    if sort_index:
        voltage_levels = net.get_voltage_levels().sort_index()
    else:
        voltage_levels = net.get_voltage_levels()
        
    all_buses_vn_kv = voltage_levels.loc[bus_df["voltage_level_id"].values]["nominal_v"].values
    sub_unique = None
    sub_unique_id = None
    if n_busbar_per_sub is not None and buses_for_sub is not None:
        # I am in a compatibility mode,
        # I need to use the same convention as grid2op
        # for the buses labelling
        bus_df = bus_df.sort_values("voltage_level_id", kind="stable")
        sub_unique = bus_df["voltage_level_id"].unique()
        nb_sub_unique = sub_unique.shape[0]
        sub_unique_id = np.arange(nb_sub_unique)
        bus_per_sub =  bus_df[["voltage_level_id", "name"]].groupby("voltage_level_id").count()
        if (bus_per_sub["name"] > n_busbar_per_sub).any():
            max_bb = bus_per_sub["name"].max()
            raise RuntimeError(f"Impossible configuration: we found a substation with {max_bb} "
                               f"while asking for {n_busbar_per_sub}. We cannot load a grid with these "
                               f"kwargs. If you use LightSimBackend, you need to change `loader_kwargs` "
                               f"and especially the `n_busbar_per_sub` to be >= {max_bb}")
        bus_local_id = np.concatenate([np.arange(el) for el in bus_per_sub.values])
        sub_id_duplicate = np.repeat(sub_unique_id, bus_per_sub.values.ravel())
        bus_global_id = bus_local_id * nb_sub_unique + sub_id_duplicate
        bus_df["bus_id"] = bus_global_id
        all_buses_vn_kv = 1. * voltage_levels["nominal_v"].values
        ls_to_orig = np.zeros(n_busbar_per_sub * nb_sub_unique, dtype=int) - 1
        ls_to_orig[bus_df["bus_id"].values] = np.arange(bus_df.shape[0])
        n_sub = nb_sub_unique
        n_bb_per_sub = n_busbar_per_sub
        bus_df = bus_df.sort_index()
    else:
        if buses_for_sub is not None:
            raise NotImplementedError("This is not implemented at the moment")
        # retrieve the labels from the buses in the original grid
        bus_df["bus_id"] = -1
        bus_df.loc[bus_df_orig.index, "bus_id"] = np.arange(bus_df.shape[0]) 
        bus_df = bus_df.sort_values("bus_id")
        ls_to_orig = 1 * bus_df["bus_id"].values
        
        n_sub = bus_df.shape[0]
        n_bb_per_sub = None
        if n_busbar_per_sub is None:
            warnings.warn("You should avoid using this function without at least `buses_for_sub` or `n_busbar_per_sub`. "
                          "Setting automatically n_busbar_per_sub=1")
            n_busbar_per_sub = 1
            
        # used to be done in the Backend previously, now we do it here instead
        bus_init = 1. * all_buses_vn_kv
        bus_doubled = np.concatenate([bus_init for _ in range(n_busbar_per_sub)])
        ls_to_orig = np.concatenate((ls_to_orig, np.full((n_busbar_per_sub - 1) * n_sub, fill_value=-1, dtype=ls_to_orig.dtype)))
        
    all_buses_vn_kv = np.concatenate([all_buses_vn_kv for _ in range(n_busbar_per_sub)])
    model.set_sn_mva(sn_mva)
    model.set_init_vm_pu(1.06)
    model.init_bus(all_buses_vn_kv,
                   0, 0  # unused
                   )
    model._ls_to_orig = ls_to_orig
    model.set_n_sub(n_sub)
    if n_bb_per_sub is not None:
        model._max_nb_bus_per_sub = n_busbar_per_sub
    
    # do the generators
    if sort_index:
        df_gen = net.get_generators().sort_index()
    else:
        df_gen = net.get_generators()
        
    # to handle encoding in 32 bits and overflow when "splitting" the Q values among 
    min_float_value = np.finfo(np.float32).min * 1e-4 + 1.
    max_float_value = np.finfo(np.float32).max * 1e-4 + 1.
    min_q_aux = 1. * df_gen["min_q"].values
    too_small = min_q_aux < min_float_value
    min_q_aux[too_small] = min_float_value
    min_q = min_q_aux.astype(np.float32)
    
    max_q_aux = 1. * df_gen["max_q"].values
    too_big = np.abs(max_q_aux) > max_float_value
    max_q_aux[too_big] = np.sign(max_q_aux[too_big]) * max_float_value
    max_q = max_q_aux.astype(np.float32)
    min_q[~np.isfinite(min_q)] = min_float_value
    max_q[~np.isfinite(max_q)] = max_float_value
    gen_bus, gen_disco = _aux_get_bus(bus_df, df_gen)

    # dirty fix for when regulating elements are not the same
    bus_reg = copy.deepcopy(df_gen["regulated_element_id"].values)
    vl_reg = copy.deepcopy(df_gen["voltage_level_id"].values)
    mask_ref_bbs = bus_reg != df_gen.index
    
    bbs_df = net.get_busbar_sections().copy()
    if not (np.isin(bus_reg[mask_ref_bbs], bbs_df.index)).all():
        raise RuntimeError("At least some generator are in 'remote control' mode "
                           "and does not control a busbar section, this is not supported "
                           "at the moment.")
    vl_reg[mask_ref_bbs] = bbs_df.loc[bus_reg[mask_ref_bbs], "voltage_level_id"].values
    model.init_generators_full(df_gen["target_p"].values,
                            #    df_gen["target_v"].values / voltage_levels.loc[df_gen["voltage_level_id"].values]["nominal_v"].values,
                               df_gen["target_v"].values / voltage_levels.loc[vl_reg]["nominal_v"].values,
                               df_gen["target_q"].values,
                               df_gen["voltage_regulator_on"].values,
                               min_q,
                               max_q,
                               gen_bus
                               )
    for gen_id, is_disco in enumerate(gen_disco):
        if is_disco:
            model.deactivate_gen(gen_id)
    model.set_gen_names(df_gen.index)   
    
    # for loads
    if sort_index:
        df_load = net.get_loads().sort_index()
    else:
        df_load = net.get_loads()
    load_bus, load_disco = _aux_get_bus(bus_df, df_load)
    model.init_loads(df_load["p0"].values,
                     df_load["q0"].values,
                     load_bus
                     )
    for load_id, is_disco in enumerate(load_disco):
        if is_disco:
            model.deactivate_load(load_id)
    model.set_load_names(df_load.index)
    
    # for lines
    if sort_index:
        df_line = net.get_lines().sort_index()
    else:
        df_line = net.get_lines()
    # per unit
    if net_pu is None:
        try:
            net_pu = copy.deepcopy(net)
            net_pu.per_unit = True
        except Exception as exc_:
            from pypowsybl.network import PerUnitView
            net_pu = PerUnitView(net)
            warnings.warn("The `PerUnitView` (python side) is less efficient and less "
                        "tested that the equivalent java class. Please upgrade pypowsybl version")
    df_line_pu = net_pu.get_lines().loc[df_line.index]
    # branch_from_kv = voltage_levels.loc[df_line["voltage_level1_id"].values]["nominal_v"].values
    # branch_to_kv = voltage_levels.loc[df_line["voltage_level2_id"].values]["nominal_v"].values
    
    # only valid for lines with same voltages at both side...
    # # branch_from_pu = branch_from_kv * branch_from_kv / sn_mva
    # # line_r = df_line["r"].values / branch_from_pu
    # # line_x = df_line["x"].values / branch_from_pu  
    # # line_h_or = (1j*df_line["g1"].values + df_line["b1"].values) * branch_from_pu
    # # line_h_ex = (1j*df_line["g2"].values + df_line["b2"].values) * branch_from_pu
    # real per unit conversion 
    # see https://github.com/powsybl/pypowsybl/issues/642
    # see https://github.com/powsybl/powsybl-core/blob/266442cbbd84f630acf786018618eaa3d496c6ba/ieee-cdf/ieee-cdf-converter/src/main/java/com/powsybl/ieeecdf/converter/IeeeCdfImporter.java#L347
    # for right formula
    # v1 = branch_from_kv
    # v2 = branch_to_kv
    # line_r = sn_mva *  df_line["r"].values / v1 / v2
    # line_x = sn_mva *  df_line["x"].values / v1 / v2
    # tmp_ = np.reciprocal(df_line["r"].values + 1j*df_line["x"].values)
    # b1 = df_line["b1"].values * v1*v1/sn_mva + (v1-v2)*tmp_.imag*v1/sn_mva
    # b2 = df_line["b2"].values * v2*v2/sn_mva + (v2-v1)*tmp_.imag*v2/sn_mva
    # g1 = df_line["g1"].values * v1*v1/sn_mva + (v1-v2)*tmp_.real*v1/sn_mva
    # g2 = df_line["g2"].values * v2*v2/sn_mva + (v2-v1)*tmp_.real*v2/sn_mva
    # line_h_or = (g1 + 1j * b1)
    # line_h_ex = (g2 + 1j * b2)
    line_r = df_line_pu["r"].values
    line_x = df_line_pu["x"].values
    line_h_or = (df_line_pu["g1"].values + 1j * df_line_pu["b1"].values)
    line_h_ex = (df_line_pu["g2"].values + 1j * df_line_pu["b2"].values)
    lor_bus, lor_disco = _aux_get_bus(bus_df, df_line, conn_key="connected1", bus_key="bus1_id")
    lex_bus, lex_disco = _aux_get_bus(bus_df, df_line, conn_key="connected2", bus_key="bus2_id")
    model.init_powerlines_full(line_r,
                               line_x,
                               line_h_or,
                               line_h_ex,
                               lor_bus,
                               lex_bus
                              )
    for line_id, (is_or_disc, is_ex_disc) in enumerate(zip(lor_disco, lex_disco)):
        if is_or_disc or is_ex_disc:
            model.deactivate_powerline(line_id)
    model.set_line_names(df_line.index)   
            
    # for trafo
    # I extract trafo with `all_attributes=True` so that I have access to the
    # `rho`
    if sort_index:
        df_trafo = net.get_2_windings_transformers(all_attributes=True).sort_index()
    else:
        df_trafo = net.get_2_windings_transformers(all_attributes=True)
        
    df_trafo_pu = net_pu.get_2_windings_transformers(all_attributes=True).loc[df_trafo.index]
    ratio_tap_changer = net_pu.get_ratio_tap_changers()
    if net.get_phase_tap_changers().shape[0] > 0:
        raise RuntimeError("Phase tap changer are not handled by the pypowsybl converter (but they are by lightsim2grid)")
    # phase_tap_changer = net_pu.get_phase_tap_changers()
    
    shift_ = np.zeros(df_trafo.shape[0])
    tap_position = 1.0 * shift_
    is_tap_hv_side = np.zeros(df_trafo.shape[0], dtype=bool)  # TODO
    
    # per unit
    # trafo_from_kv = voltage_levels.loc[df_trafo["voltage_level1_id"].values]["nominal_v"].values
    # trafo_to_kv = voltage_levels.loc[df_trafo["voltage_level2_id"].values]["nominal_v"].values
    # trafo_to_pu = trafo_to_kv * trafo_to_kv / sn_mva
    # tap
    # tap_step_pct = (df_trafo["rated_u1"] / trafo_from_kv - 1.) * 100.
    # has_tap = tap_step_pct != 0.
    # tap_pos[has_tap] += 1
    # tap_step_pct[~has_tap] = 1.0  # or any other values...
    # trafo_r = df_trafo["r"].values / trafo_to_pu
    # trafo_x = df_trafo["x"].values / trafo_to_pu
    # trafo_h = (df_trafo["g"].values + 1j * df_trafo["b"].values) * trafo_to_pu
    trafo_r = df_trafo_pu["r"].values
    trafo_x = df_trafo_pu["x"].values
    trafo_h = (df_trafo_pu["g"].values + 1j * df_trafo_pu["b"].values)
    
    # now get the ratio    
    # in lightsim2grid (cpp)
    # RealVect ratio = my_one_ + 0.01 * trafo_tap_step_pct.array() * trafo_tap_pos.array();
    # in powsybl (https://javadoc.io/doc/com.powsybl/powsybl-core/latest/com/powsybl/iidm/network/TwoWindingsTransformer.html)
    #  rho = transfo.getRatedU2() / transfo.getRatedU1()
    # * (transfo.getRatioTapChanger() != null ? transfo.getRatioTapChanger().getCurrentStep().getRho() : 1);
    # * (transfo.getPhaseTapChanger() != null ? transfo.getPhaseTapChanger().getCurrentStep().getRho() : 1);
    ratio = 1. * (df_trafo_pu["rated_u2"].values / df_trafo_pu["rated_u1"].values)
    has_r_tap_changer = np.isin(df_trafo_pu.index, ratio_tap_changer.index)
    
    if PYPOWSYBL_VER <= PP_BUG_RATIO_TAP_CHANGER:
        # bug in per unit view in both python and java
        ratio[has_r_tap_changer] = 1. * ratio_tap_changer.loc[df_trafo_pu.loc[has_r_tap_changer].index, "rho"].values
    else:
        ratio[has_r_tap_changer] = 1. * df_trafo_pu.loc[has_r_tap_changer, "rho"].values
    no_tap = ratio == 1.
    tap_neg = ratio < 1. 
    tap_positive = ratio > 1. 
    tap_step_pct = 1. * ratio
    tap_step_pct[tap_positive] -= 1.
    tap_step_pct[tap_positive] *= 100.
    tap_step_pct[tap_neg] = (ratio[tap_neg] - 1.)*100.
    tap_step_pct[no_tap] = 100.
    tap_position[tap_positive] += 1
    tap_position[tap_neg] += 1
    
    tor_bus, tor_disco = _aux_get_bus(bus_df, df_trafo, conn_key="connected1", bus_key="bus1_id")
    tex_bus, tex_disco = _aux_get_bus(bus_df, df_trafo, conn_key="connected2", bus_key="bus2_id")
    model.init_trafo(trafo_r,
                     trafo_x,
                     trafo_h,
                     tap_step_pct,
                     tap_position,
                     shift_,
                     is_tap_hv_side,
                     tor_bus, # TODO do I need to change hv / lv
                     tex_bus)    
    for t_id, (is_or_disc, is_ex_disc) in enumerate(zip(tor_disco, tex_disco)):
        if is_or_disc or is_ex_disc:
            model.deactivate_trafo(t_id)
    model.set_trafo_names(df_trafo.index)
    
    # for shunt
    if sort_index:
        df_shunt = net.get_shunt_compensators().sort_index()
    else:
        df_shunt = net.get_shunt_compensators()
        
    sh_bus, sh_disco = _aux_get_bus(bus_df, df_shunt)    
    shunt_kv = voltage_levels.loc[df_shunt["voltage_level_id"].values]["nominal_v"].values
    model.init_shunt(-df_shunt["g"].values * shunt_kv**2,
                     -df_shunt["b"].values * shunt_kv**2,
                     sh_bus
                    )
    for shunt_id, disco in enumerate(sh_disco):
        if disco:
           model.deactivate_shunt(shunt_id) 
    model.set_shunt_names(df_shunt.index)
           
    # for hvdc (TODO not tested yet)
    if sort_index:
        df_dc = net.get_hvdc_lines().sort_index()
        df_sations = net.get_vsc_converter_stations().sort_index()
    else:
        df_dc = net.get_hvdc_lines()
        df_sations = net.get_vsc_converter_stations()
    # bus_from_id = df_sations.loc[df_dc["converter_station1_id"].values]["bus_id"].values
    # bus_to_id = df_sations.loc[df_dc["converter_station2_id"].values]["bus_id"].values
    hvdc_bus_from_id, hvdc_from_disco = _aux_get_bus(bus_df, df_sations.loc[df_dc["converter_station1_id"].values]) 
    hvdc_bus_to_id, hvdc_to_disco = _aux_get_bus(bus_df, df_sations.loc[df_dc["converter_station2_id"].values]) 
    loss_percent = np.zeros(df_dc.shape[0])  # TODO 
    loss_mw = np.zeros(df_dc.shape[0])  # TODO
    model.init_dclines(hvdc_bus_from_id,
                       hvdc_bus_to_id,
                       df_dc["target_p"].values,
                       loss_percent,
                       loss_mw,
                       voltage_levels.loc[df_sations.loc[df_dc["converter_station1_id"].values]["voltage_level_id"].values]["nominal_v"].values,
                       voltage_levels.loc[df_sations.loc[df_dc["converter_station2_id"].values]["voltage_level_id"].values]["nominal_v"].values,
                       df_sations.loc[df_dc["converter_station1_id"].values]["min_q"].values,
                       df_sations.loc[df_dc["converter_station1_id"].values]["max_q"].values,
                       df_sations.loc[df_dc["converter_station2_id"].values]["min_q"].values,
                       df_sations.loc[df_dc["converter_station2_id"].values]["max_q"].values
                       )
    # TODO will probably not work !
    for hvdc_id, (is_or_disc, is_ex_disc) in enumerate(zip(hvdc_from_disco, hvdc_to_disco)):
        if is_or_disc or is_ex_disc:
            model.deactivate_hvdc(hvdc_id)
    model.set_dcline_names(df_sations.index)
                
    # storage units  (TODO not tested yet)
    if sort_index:
        df_batt = net.get_batteries().sort_index()
    else:
        df_batt = net.get_batteries()
    batt_bus, batt_disco = _aux_get_bus(bus_df, df_batt)
    model.init_storages(df_batt["target_p"].values,
                        df_batt["target_q"].values,
                        batt_bus
                        )
    for batt_id, disco in enumerate(batt_disco):
        if disco:
           model.deactivate_storage(batt_id) 
    model.set_storage_names(df_batt.index)

    # TODO dist slack
    if gen_slack_id is None and slack_bus_id is None:
        # if nothing is given, by default I assign a slack bus to a bus where a lot of lines are connected
        # quite central in the grid
        bus_id, gen_id = model.assign_slack_to_most_connected()
    elif gen_slack_id is not None:
        if slack_bus_id is not None:
            raise RuntimeError(f"You provided both gen_slack_id and slack_bus_id which is not possible.")
        
        if isinstance(gen_slack_id, str):
            gen_slack_id_int = int((df_gen.index == gen_slack_id).nonzero()[0][0])
        else:
            try:
                gen_slack_id_int = int(gen_slack_id)
            except Exception as exc_:
                raise RuntimeError("'slack_bus_id' should be either an int or a generator names") from exc_
            if gen_slack_id_int != gen_slack_id:
                raise RuntimeError("'slack_bus_id' should be either an int or a generator names")
        model.add_gen_slackbus(gen_slack_id_int, 1.)
    elif slack_bus_id is not None:
        gen_bus = np.array([el.bus_id for el in model.get_generators()])
        gen_is_conn_slack = gen_bus == model._orig_to_ls[slack_bus_id]
        nb_conn = gen_is_conn_slack.sum()
        if nb_conn == 0:
            raise RuntimeError(f"There is no generator connected to bus {slack_bus_id}. It cannot be the slack")
        for gen_id, is_slack in enumerate(gen_is_conn_slack):
            if is_slack:
                model.add_gen_slackbus(gen_id, 1. / nb_conn)    
    else:
        raise RuntimeError("You need to provide at least one slack with `gen_slack_id` or `slack_bus_id`")
    
    # TODO
    # sgen => regular gen (from net.get_generators()) with voltage_regulator off TODO 
    
    # TODO checks
    # no 3windings trafo and other exotic stuff
    if net.get_phase_tap_changers().shape[0] > 0:
        warnings.warn("There are tap changers in the iidm grid which are not taken "
                      "into account in the lightsim2grid at the moment. "
                      "NB: lightsim2grid gridmodel can handle tap changer, it is just not "
                      "handled by the 'from_pypowsybl` function at the moment.")
        
    # and now deactivate all elements and nodes not in the main component
    if only_main_component:
        model.consider_only_main_component()
    model.init_bus_status()  # automatically disconnect non connected buses
    if not return_sub_id:
        # for backward compatibility
        return model
    else:
        # voltage_level_id is kind of what I call "substation" in grid2op
        if sub_unique is None:
            vl_unique = bus_df["voltage_level_id"].unique()
            sub_df = pd.DataFrame(index=np.sort(vl_unique), data={"sub_id": np.arange(vl_unique.size)})
        else:
            # already computed before when innitializing the buses and substations
            # I don't do it again to ensure consistency
            sub_df = pd.DataFrame(index=sub_unique, data={"sub_id": sub_unique_id})
        buses_sub_id = pd.merge(left=bus_df, right=sub_df, how="left", left_on="voltage_level_id", right_index=True)[["bus_id", "sub_id"]]
        gen_sub = pd.merge(left=df_gen, right=sub_df, how="left", left_on="voltage_level_id", right_index=True)[["sub_id"]]
        load_sub = pd.merge(left=df_load, right=sub_df, how="left", left_on="voltage_level_id", right_index=True)[["sub_id"]]
        lor_sub = pd.merge(left=df_line, right=sub_df, how="left", left_on="voltage_level1_id", right_index=True)[["sub_id"]]
        lex_sub = pd.merge(left=df_line, right=sub_df, how="left", left_on="voltage_level2_id", right_index=True)[["sub_id"]]
        tor_sub = pd.merge(left=df_trafo, right=sub_df, how="left", left_on="voltage_level1_id", right_index=True)[["sub_id"]]
        tex_sub = pd.merge(left=df_trafo, right=sub_df, how="left", left_on="voltage_level2_id", right_index=True)[["sub_id"]]
        batt_sub = pd.merge(left=df_batt, right=sub_df, how="left", left_on="voltage_level_id", right_index=True)[["sub_id"]]
        sh_sub = pd.merge(left=df_shunt, right=sub_df, how="left", left_on="voltage_level_id", right_index=True)[["sub_id"]]
        hvdc_vl_info = pd.DataFrame(index=df_dc.index,
                                    data={"voltage_level1_id": df_sations.loc[df_dc["converter_station1_id"].values]["voltage_level_id"].values,
                                          "voltage_level2_id": df_sations.loc[df_dc["converter_station2_id"].values]["voltage_level_id"].values
                                          })
        hvdc_sub_from_id = pd.merge(left=hvdc_vl_info, right=sub_df, how="left", left_on="voltage_level1_id", right_index=True)[["sub_id"]]
        hvdc_sub_to_id = pd.merge(left=hvdc_vl_info, right=sub_df, how="left", left_on="voltage_level2_id", right_index=True)[["sub_id"]]
        return model, (buses_sub_id, gen_sub, load_sub, (lor_sub, tor_sub), (lex_sub, tex_sub), batt_sub, sh_sub, hvdc_sub_from_id, hvdc_sub_to_id)



# Copyright (c) 2020-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import warnings
import numpy as np
from ._pp_bus_to_ls_bus import pp_bus_to_ls

SOME_KIND_OF_INF_FOR_PMIN_PMAX = 99999.


def _aux_add_sgen(model, pp_net, pp_to_ls):
    """
    Add the static generators  (=PQ generators) of the pp_net into the lightsim2grid "model"

    Parameters
    ----------
    model
    pp_net

    Returns
    -------

    """
    if "parallel" in pp_net.sgen and np.any(pp_net.sgen["parallel"].values != 1):
        raise RuntimeError("Cannot handle 'parallel' sgen columns. Please duplicate the rows if that is the case. "
                           "Some pp_net.sgen[\"parallel\"] != 1 it is not handled by lightsim yet.")

    if pp_net.sgen.shape[0] == 0:
        # nothing to do if no static generators
        return

    if "min_p_mw" in pp_net.sgen:
        min_p_mw = pp_net.sgen["min_p_mw"].values
    else:
        min_p_mw = np.zeros(pp_net.sgen.shape[0]) - SOME_KIND_OF_INF_FOR_PMIN_PMAX
    if np.any(~np.isfinite(min_p_mw)):
        warnings.warn("There were some Nan in the pp_net.sgen[\"min_p_mw\"], they have been replaced by 0")
    min_p_mw[~np.isfinite(min_p_mw)] = 0.

    if "max_p_mw" in pp_net.sgen:
        max_p_mw = pp_net.sgen["max_p_mw"].values
    else:
        max_p_mw = np.zeros(pp_net.sgen.shape[0]) + SOME_KIND_OF_INF_FOR_PMIN_PMAX
    if np.any(~np.isfinite(max_p_mw)):
        warnings.warn("There were some Nan in the pp_net.sgen[\"max_p_mw\"], they have been replaced by 0")
    max_p_mw[~np.isfinite(max_p_mw)] = 0.

    if "min_q_mvar" in pp_net.sgen:
        min_q_mvar = pp_net.sgen["min_q_mvar"].values
    else:
        min_q_mvar = np.zeros(pp_net.sgen.shape[0]) - SOME_KIND_OF_INF_FOR_PMIN_PMAX
    if np.any(~np.isfinite(min_q_mvar)):
        warnings.warn("There were some Nan in the pp_net.sgen[\"min_q_mvar\"], they have been replaced by 0")
    min_q_mvar[~np.isfinite(min_q_mvar)] = 0.

    if "max_q_mvar" in pp_net.sgen:
        max_q_mvar = pp_net.sgen["max_q_mvar"].values
    else:
        max_q_mvar = np.zeros(pp_net.sgen.shape[0]) + SOME_KIND_OF_INF_FOR_PMIN_PMAX
    if np.any(~np.isfinite(max_q_mvar)):
        warnings.warn("There were some Nan in the pp_net.sgen[\"max_q_mvar\"], they have been replaced by 0")
    max_q_mvar[~np.isfinite(max_q_mvar)] = 0.

    ratio = 1.0
    if "scaling" in pp_net.sgen:
        ratio = pp_net.sgen["scaling"].values
        ratio[~np.isfinite(ratio)] = 1.0
        
    model.init_sgens(pp_net.sgen["p_mw"].values * ratio,
                     pp_net.sgen["q_mvar"].values * ratio,
                     min_p_mw,
                     max_p_mw,
                     min_q_mvar,
                     max_q_mvar,
                     pp_bus_to_ls(pp_net.sgen["bus"].values, pp_to_ls)
                     )
    for sgen_id, is_connected in enumerate(pp_net.sgen["in_service"].values):
        if not is_connected:
            # load is deactivated
            model.deactivate_sgen(sgen_id)



# Copyright (c) 2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import warnings
# Deprecated now, will be removed
warnings.warn("You are using old names. Please upgrade to SecurityAnalysisCPP > ContingencyAnalysisCPP"
              " and SecurityAnalysis > ContingencyAnalysis instead.",
              category=DeprecationWarning)
__all__ = ["SecurityAnalysisCPP"]


from lightsim2grid.contingencyAnalysis import ContingencyAnalysisCPP

try:
    from lightsim2grid.contingencyAnalysis import ContingencyAnalysis
    GRID2OP_INSTALLED = True
except ImportError as exc_:
    GRID2OP_INSTALLED = False


SecurityAnalysisCPP = ContingencyAnalysisCPP


if GRID2OP_INSTALLED:
    SecurityAnalysis = ContingencyAnalysis
    __all__.append("SecurityAnalysis")



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import setuptools
from setuptools import setup
import sys
import os
import warnings
from pybind11.setup_helpers import Pybind11Extension, build_ext


__version__ = "0.10.1.post1"
KLU_SOLVER_AVAILABLE = False

# Try to link against SuiteSparse (if available)
# check that they exist (if SuiteSparse has been built with "make")
suitesparse_path_make = os.path.abspath("./SuiteSparse")
LIBS_MAKE = ["{}/KLU/Lib/libklu.a",
             "{}/BTF/Lib/libbtf.a",
             "{}/AMD/Lib/libamd.a",
             "{}/COLAMD/Lib/libcolamd.a",
             "{}/CXSparse/Lib/libcxsparse.a",
             "{}/SuiteSparse_config/libsuitesparseconfig.a"
             ]
LIBS_MAKE = [el.format(suitesparse_path_make) for el in LIBS_MAKE]
exists_libs_make = True
for el in LIBS_MAKE:
    if not os.path.exists(el):
        exists_libs_make = False

# check that they exist (if SuiteSparse has been built with "cmake" on macos / linux or windows)
suitesparse_path_cmake = os.path.abspath("./build_cmake/built/")
for ext in ["a", "lib"]:
    LIBS_CMAKE = [f"libklu.{ext}",
                  f"libbtf.{ext}",
                  f"libamd.{ext}",
                  f"libcolamd.{ext}",
                  f"libcxsparse.{ext}"]
    if ext == "a":
        # unix like system
        LIBS_CMAKE.append(f"libsuitesparseconfig.{ext}")
    else:
        # windows like system
        LIBS_CMAKE.append(f"suitesparseconfig.{ext}")

    LIBS_CMAKE = [os.path.join(suitesparse_path_cmake, "lib", el) for el in LIBS_CMAKE]

    exists_libs_cmake = True
    for el in LIBS_CMAKE:
        if not os.path.exists(el):
            exists_libs_cmake = False
            break

    if exists_libs_cmake:
        break


if exists_libs_make:
    # you will be able to use "SuiteSparse" and the faster "KLU" linear solver
    KLU_SOLVER_AVAILABLE = True

    # include directory
    INCLUDE_suitesparse = ["{}/SuiteSparse_config",
                           "{}/CXSparse/Include",
                           "{}/AMD/Include",
                           "{}/BTF/Include",
                           "{}/COLAMD/Include",
                           "{}/KLU/Include"
                           ]
    INCLUDE_suitesparse = [el.format(suitesparse_path_make) for el in INCLUDE_suitesparse]

    # compiled libraries location
    LIBS = LIBS_MAKE
elif exists_libs_cmake:
    # you will be able to use "SuiteSparse" and the faster "KLU" linear solver
    KLU_SOLVER_AVAILABLE = True

    # include directory
    INCLUDE_suitesparse = [os.path.join(suitesparse_path_cmake, "include", "suitesparse")]

    # compiled libraries location
    LIBS = LIBS_CMAKE
else:
    # SuiteSparse, and in particular the KLU linear solver is not available.
    # we'll use a default solver (a bit slower)
    LIBS = []
    INCLUDE_suitesparse = []
    warnings.warn("SuiteSparse is not available on your system, or has not been compiled. The faster "
                  "\"KLU\" linear algebra solver will not be available. The \"SparseLU\" solver will however "
                  "be available, which is maybe ~30% slower than \"KLU\". If you are using grid2op there "
                  "will still be a huge benefit.")

INCLUDE = ["src"] + INCLUDE_suitesparse

# now add the Eigen library (header only)
eigen_path = os.path.abspath(".")
INCLUDE.append("{}/eigen".format(eigen_path))

include_dirs = []
include_dirs += INCLUDE

# extract the version information
VERSION_MAJOR, VERSION_MEDIUM, VERSION_MINOR = __version__.split(".")[:3]

# compiler options
extra_compile_args_tmp = ["-DNDEBUG"]
IS_LINUX = False
IS_MACOS = False
IS_WINDOWS = False
if sys.platform.startswith('linux'):
    # extra_compile_args_tmp = ["-fext-numeric-literals"]
    # -fext-numeric-literals is used for definition of complex number by some version of gcc
    extra_compile_args_tmp += []
    IS_LINUX = True
elif sys.platform.startswith("darwin"):
    # extra_compile_args_tmp = ["-fsized-deallocation"]
    extra_compile_args_tmp += []
    # fix a bug in pybind11
    # https://github.com/pybind/pybind11/issues/1604
    IS_MACOS = True
elif sys.platform.startswith("win32"):
    extra_compile_args_tmp += [# otherwise windows compiler does not import "M_PI" from the math header
                               "-D_USE_MATH_DEFINES"]
    IS_WINDOWS = True

# if you have installed some BLAS or LAPACKE libraries (on ubuntu sudo apt-get install libblas-dev liblapacke-dev)
# you can also trigger their use when using eigen.
# extra_compile_args_tmp += ["-DEIGEN_USE_BLAS", "-DEIGEN_USE_LAPACKE"]
extra_compile_args = extra_compile_args_tmp
# add the version information
extra_compile_args += [f"-DVERSION_MAJOR={VERSION_MAJOR}",
                       f"-DVERSION_MEDIUM={VERSION_MEDIUM}",
                       f"-DVERSION_MINOR={VERSION_MINOR}"]
extra_compile_args += [f"-DVERSION=\"{__version__}\""]
src_files = ['src/main.cpp',
             "src/powerflow_algorithm/GaussSeidelAlgo.cpp",
             "src/powerflow_algorithm/GaussSeidelSynchAlgo.cpp",
             "src/powerflow_algorithm/BaseAlgo.cpp",
             "src/linear_solvers/SparseLUSolver.cpp",
             "src/help_fun_msg.cpp",
             "src/BaseConstants.cpp",
             "src/GridModel.cpp",
             "src/ChooseSolver.cpp",
             "src/Solvers.cpp",
             "src/Utils.cpp",
             "src/DataConverter.cpp",
             "src/batch_algorithm/BaseBatchSolverSynch.cpp",
             "src/batch_algorithm/TimeSeries.cpp",
             "src/batch_algorithm/ContingencyAnalysis.cpp",
             "src/element_container/LineContainer.cpp",
             "src/element_container/GenericContainer.cpp",
             "src/element_container/ShuntContainer.cpp",
             "src/element_container/TrafoContainer.cpp",
             "src/element_container/LoadContainer.cpp",
             "src/element_container/GeneratorContainer.cpp",
             "src/element_container/SGenContainer.cpp",
             "src/element_container/DCLineContainer.cpp"]

if KLU_SOLVER_AVAILABLE:
    src_files.append("src/linear_solvers/KLUSolver.cpp")
    extra_compile_args_tmp.append("-DKLU_SOLVER_AVAILABLE")
    print("INFO: Using KLU package")

# Try to locate the NICSLU sparse linaer solver
if "PATH_NICSLU" in os.environ:
    # user indicate the path for the NICSLU library (see https://github.com/chenxm1986/nicslu)
    # eg "/home/user/Documents/nicslu/nicslu202103/"

    path_nicslu = os.path.abspath(os.environ["PATH_NICSLU"])
    include_nicslu = True
    # check for appropriate license
    if not os.path.exists(path_nicslu):
        print(f"WARNING: nothing for NICSLU at at: {path_nicslu}")
        include_nicslu = False
    license_path = os.path.join(path_nicslu, "license")
    if include_nicslu and not os.path.exists(license_path):
        # license not located at the right directory
        print(f"WARNING: no license path found for NICSLU at: {license_path}")
        include_nicslu = False
    license_file = os.path.join(license_path, "nicslu.lic")
    if include_nicslu and not os.path.exists(license_file):
        # no license found
        print(f"WARNING: no license file is found for NICSLU at: {license_file}")
        include_nicslu = False
    libnicslu_path = None
    if include_nicslu:
        if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
            libnicslu_path = os.path.join(path_nicslu, "linux/lib_centos6_x64_gcc482_fma/int32/libnicslu.so")
            if not os.path.exists(libnicslu_path):
                print(f"WARNING: cannot locate the NICSLU shared object that should be at: {libnicslu_path}")
                include_nicslu = False
                libnicslu_path = None
        elif sys.platform.startswith("win"):
            libnicslu_path = os.path.join(path_nicslu, "windows/lib_win7_x64_fma/int32/nicslu.lib")
            if not os.path.exists(libnicslu_path):
                print(f"WARNING: cannot locate the NICSLU shared object that should be at: {libnicslu_path}")
                include_nicslu = False
                libnicslu_path = None
        else:
            print(f"WARNING: NICSLU can only be added when using linux, darwin (MacOS) or win (Windows) python version, you are using {sys.platform}")
            include_nicslu = False
            libnicslu_path = None

    if include_nicslu and libnicslu_path is not None:
        LIBS.append(os.path.join(path_nicslu, libnicslu_path))
        include_dirs.append(os.path.join(path_nicslu, "include"))
        src_files.append("src/linear_solvers/NICSLUSolver.cpp")
        extra_compile_args.append("-DNICSLU_SOLVER_AVAILABLE")
        extra_compile_args += [f"-DNICSLU_PATH=\"{os.path.join(path_nicslu, libnicslu_path)}\""]
        print("INFO: Using NICSLU package")

# Try to locate the CKTSO sparse linear solver
if "PATH_CKTSO" in os.environ:
    # user indicate the path for the CKTSO library (see https://github.com/chenxm1986/cktso)
    # eg "/home/user/Documents/cktso/"
    
    path_cktso = os.path.abspath(os.environ["PATH_CKTSO"])
    include_cktso = True
    # check for appropriate license
    if not os.path.exists(path_cktso):
        print(f"WARNING: nothing for CKTSO at at: {path_cktso}")
        include_cktso = False
    license_path = os.path.join(path_cktso, "license")
    if include_cktso and not os.path.exists(license_path):
        # license not located at the right directory
        print(f"WARNING: no license path found for CKTSO at: {license_path}")
        include_cktso = False
    license_file = os.path.join(license_path, "cktso.lic")
    if include_cktso and not os.path.exists(license_file):
        # no license found
        print(f"WARNING: no license file is found for CKTSO at: {license_file}")
        include_cktso = False
    libcktso_path = None
    if include_cktso:
        if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
            libcktso_path = os.path.join(path_cktso, "ubuntu1804_x64_gcc750/libcktso.so")  # TODO CHANGE THAT !
            if not os.path.exists(libcktso_path):
                print(f"WARNING: cannot locate the CKTSO shared object that should be at: {libcktso_path}")
                include_cktso = False
                libcktso_path = None
        elif sys.platform.startswith("win"):
            libcktso_path = os.path.join(path_cktso, "win7_x64/cktso.lib")  # TODO CHANGE THAT !
            if not os.path.exists(libcktso_path):
                print(f"WARNING: cannot locate the CKTSO shared object that should be at: {libcktso_path}")
                include_cktso = False
                libcktso_path = None
        else:
            print(f"WARNING: CKTSO can only be added when using linux, darwin (MacOS) or win (Windows) python version, you are using {sys.platform}")
            include_cktso = False
            libcktso_path = None

    if include_cktso and libcktso_path is not None:
        LIBS.append(os.path.join(path_cktso, libcktso_path))
        include_dirs.append(os.path.join(path_cktso, "include"))
        src_files.append("src/linear_solvers/CKTSOSolver.cpp")
        extra_compile_args.append("-DCKTSO_SOLVER_AVAILABLE")
        extra_compile_args += [f"-DCKTSO_PATH=\"{os.path.join(path_cktso, libcktso_path)}\""]
        print("INFO: Using CKTSO package")
        
        
if "__COUT_TIMES" in os.environ:
    # to add extra info in cout for the computation times, we do not recommend to use it !
    if os.environ["__COUT_TIMES"] == "1":
        extra_compile_args.append("-D__COUT_TIMES")
        print("WARNING: Using the \"cout times\" compatibility mode, do not use the generated package outside of testing !")

if "__COMPILE_MARCHNATIVE" in os.environ:
    if os.environ["__COMPILE_MARCHNATIVE"] == "1":
        extra_compile_args.append("-march=native")
        extra_compile_args.append("-D__COMPILE_MARCHNATIVE")
        print("INFO: Using \"-march=native\" compiler flag")

# $Env:_READ_THE_DOCS = "1" in powershell
if "_READ_THE_DOCS" in os.environ:
    # generation is made on readthedocs.org for documentation, everything must be added, even though some packages will
    # not be available (eg KLU, NICSLU, etc.)

    if os.environ["_READ_THE_DOCS"] == "1":
        extra_compile_args.append("-D_READ_THE_DOCS")
        print("WARNING: Using the \"read the docs\" compatibility mode, do not use the package for something else than generating documentation.")
        
if False:
    path_iidm = ""
    lib_iidm = [os.path.join(path_iidm, "lib", "libiidm.a")]

    src_files.append("src/IIDMConverter.cpp")
    extra_compile_args_tmp.append("-DIIDM_CONVERTER_AVAILABLE")

    include_dirs.append(os.path.join(path_iidm, "include"))
    include_dirs.append("/usr/include/libxml2/")  # for libxml2
    LIBS += lib_iidm

if "__O3_OPTIM" in os.environ:
    # to add extra info in cout for the computation times, we do not recommend to use it !
    if os.environ["__O3_OPTIM"] == "1":
        if IS_LINUX or IS_MACOS:
            extra_compile_args.append("-O3")
            print("INFO: Using \"-O3\" compiler flag")
        elif IS_WINDOWS:
            extra_compile_args.append("/O2")
        extra_compile_args.append("-D__O3_OPTIM")

ext_modules = [
    Pybind11Extension(
        'lightsim2grid_cpp',
        src_files,
        include_dirs=include_dirs,
        extra_objects=LIBS,
        extra_compile_args=extra_compile_args
    )
]

# before pandapower 2.8 scipy was forced to be <= 1.6 which does not work with
# python 3.10+
req_pkgs = [
    "setuptools",
    "pip",
    "pybind11",
    "scipy",
    "numpy",
    "packaging", 
        # "pandapower" if sys.version_info < (3, 10) else "pandapower>=2.8",
        # "pytest",  # for pandapower see https://github.com/e2nIEE/pandapower/issues/1988
    ]
if sys.version_info < (3, 11):
    req_pkgs.append("typing_extensions")
if sys.version_info.major == 3 and sys.version_info.minor <= 7:
    # typing "Literal" not available on python 3.7
    req_pkgs.append("typing_extensions")
    # do not use pandapower 2.12 (broken on python 3.7 
    # see https://github.com/e2nIEE/pandapower/issues/1985
    req_pkgs[0] = "pandapower>=2.2.2,<2.12"

pkgs = {
    "required": req_pkgs,
    "extras": {
        "docs": [
            "numpydoc>=0.9.2",
            "sphinx>=2.4.4,<7",
            "sphinx-rtd-theme>=0.4.3",
            "sphinxcontrib-trio>=1.1.0",
            "autodocsumm>=0.1.13",
            "grid2op>=1.6.4",
            "recommonmark",
            "pypowsybl"
        ],
        "benchmark": [
            "tabulate",
            "grid2op>=1.6.4",
            "numpy>=1.20",
            "distro",
            "py-cpuinfo"
        ],
        "recommended": [
            "grid2op>=1.6.4",
            "numba"
        ],
        "test": [
            "grid2op>=1.6.4",
            "numba",
            "pandapower>=2.14.0",  # interface changed for pandapower, not backward compatible
            "pypowsybl"
        ]
    }
}

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='LightSim2Grid',
      version=__version__,
      author='Benjamin Donnot',
      author_email='benjamin.donnot@rte-france.com',
      url='https://github.com/grid2op/lightsim2grid/',
      description='LightSim2Grid implements a c++ backend targeting the Grid2Op platform.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      ext_modules=ext_modules,
      install_requires=pkgs["required"],
      extras_require=pkgs["extras"],
      # setup_requires=['pybind11>=2.4'],  # in the pyproject.toml directly now
      cmdclass={'build_ext': build_ext},
      zip_safe=False,
      license='MPL 2.0',
      platforms=["Windows", "Linux", "Mac OS-X", "Unix"],
      packages=setuptools.find_packages(),
      keywords='pandapower powergrid simulator KLU Eigen c++',
      classifiers=[
            'Development Status :: 4 - Beta',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "Natural Language :: English"
      ]
      )



# Copyright (c) 2020-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import numpy as np

from ._pp_bus_to_ls_bus import pp_bus_to_ls

def _aux_add_load(model, pp_net, pp_to_ls):
    """
    Add the load of the pp_net into the lightsim2grid "model"

    Parameters
    ----------
    model
    pp_net

    Returns
    -------

    """
    if "parallel" in pp_net.load and np.any(pp_net.load["parallel"].values != 1):
        raise RuntimeError("Cannot handle 'parallel' load columns. Please duplicate the rows if that is the case. "
                           "Some pp_net.load[\"parallel\"] != 1 it is not handled by lightsim yet.")

    ratio = 1.0
    if "scaling" in pp_net.load:
        ratio = pp_net.load["scaling"].values
        ratio[~np.isfinite(ratio)] = 1.0
        
    model.init_loads(pp_net.load["p_mw"].values * ratio,
                     pp_net.load["q_mvar"].values * ratio,
                     pp_bus_to_ls(pp_net.load["bus"].values, pp_to_ls)
                     )
    for load_id, is_connected in enumerate(pp_net.load["in_service"].values):
        if not is_connected:
            # load is deactivated
            model.deactivate_load(load_id)



# Copyright (c) 2024, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

__all__ = ["klu_solver_available",
           "nicslu_solver_available",
           "cktso_solver_available",
           "compiled_march_native",
           "compiled_o3_optim",
           "version",
           "lightsim2grid_lib"]

from lightsim2grid_cpp import klu_solver_available
from lightsim2grid_cpp import nicslu_solver_available
from lightsim2grid_cpp import cktso_solver_available
from lightsim2grid_cpp import compiled_march_native
from lightsim2grid_cpp import compiled_o3_optim
from lightsim2grid_cpp import version
from lightsim2grid_cpp import __file__ as lightsim2grid_lib

try:
    from lightsim2grid_cpp import nicslu_lib
    __all__.append("nicslu_lib")
except ImportError :
    # NICSLU linear solver is not available
    pass

try:
    from lightsim2grid_cpp import cktso_lib
    __all__.append("cktso_lib")
except ImportError :
    # CKTSO linear solver is not available
    pass



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of LightSim2grid, LightSim2grid implements a c++ backend targeting the Grid2Op platform.

import numpy as np
from scipy import sparse

from lightsim2grid.solver import DCSolver

try:
    from lightsim2grid.solver import KLUDCSolver
    KLU_solver_available = True
except ImportError:
    KLU_solver_available = False


def _isolate_slack_ids(Sbus, pv, pq):
    # extract the slack bus
    ref = set(np.arange(Sbus.shape[0])) - set(pv) - set(pq)
    ref = np.array(list(ref))
    # build the slack weights
    slack_weights = np.zeros(Sbus.shape[0])
    slack_weights[ref] = 1.0 / ref.shape[0]
    return ref, slack_weights


def _get_valid_solver(options, Bbus):
    # initialize the solver
    # TODO have that in options maybe (can use GaussSeidel, and NR with KLU -faster- or SparseLU)
    solver = KLUDCSolver() if KLU_solver_available else DCSolver()

    if not sparse.isspmatrix_csc(Bbus):
        Bbus = sparse.csc_matrix(Bbus)

    if not Bbus.has_canonical_format:
        Bbus.sum_duplicates()
        if not Bbus.has_canonical_format:
            raise RuntimeError("Your matrix should be in a canonical format. See "
                            "https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csc_matrix.has_canonical_format.html"
                            " for more information.")
    
    return solver


def dcpf(B, Pbus, Va0, ref, pv, pq):
    """
    Implementation (using lightsim2grid) of the `dcpf` function of pandapower
    that allows to perform DC powerflow.
    
    It is meant to be used from inside pandapower and not directly.
    """
    # initialize the solver and perform some sanity checks
    solver = _get_valid_solver(None, B)
    
    # do the newton raphson algorithm
    slack_weights = np.ones(len(ref), dtype=float)
    slack_weights /= slack_weights.sum()
    solver.solve(B, np.cos(Va0) + 1j * np.sin(Va0), Pbus, ref, slack_weights, pv, pq, 1, 1e-8)
    # extract the results
    Va = solver.get_Va()
    return Va
